/* cpu4 simulator version 1.0

simulate cpu2004 instruction set


implement Big endian
jsr 1000  print ac
jsr 1001  stop the simulator


input takes S-format file

Prabhas Chongstitvatana
8th September 1999
20 July 2004

*/

#include <stdio.h>
#include <stdlib.h>

#include "opcode.h"

#define  MXMEM	1000

#define  VF 4   /* flag affected */
#define  XF 5   /* don't affected */

#define  PRINT  1000
#define  STOP	1001

typedef struct { char name[5]; char gr; } optype;

/* gr: 0 IH, 1 IM+DI, 2 DI, 3 RE+DI */
optype optable[] = {
	"LDS",1, "LDA",1, "PSH",0, "STA",2, "JMP",3, "JPZ",3,
	"JPC",3, "JSR",3, "AND",1, "ORA",1, "POP",0, "XOR",1,
	"ADD",1, "RTS",0, "ROL",0, "ROR",0
};

char M[MXMEM];
char Cf, Zf;	/* flags : carry, zero */
int  ac, pc, sp;

int op, trace;

void error( char *s ){
  printf("error %s at pc = %d op = %d\n", s,pc,op);
  exit(0);
}

int get2( int ads){
  int v;
  uchar c1, c2;
  c1 = M[ads];
  c2 = M[ads+1];
  v = (256*c1 + c2) & 0x0ffff; /* 16-bit */
  return v;
}
/*
void put2( int ads, int value){
  M[ads] = (value >> 8) & 255;
  M[ads+1] = value & 255;
}
*/
void setZf(){
  Zf = (ac == 0) ? 1 : 0;
}

void setCf(){
  ac &= 511;
  Cf = (ac > 255) ? 1 : 0;
  ac &= 255;
}
/* VF affect flag, XF unaffect */
void setflag(char carry, char zero){
  switch(carry){
  case 0: Cf = 0; break;
  case 1: Cf = 1; break;
  case VF: setCf();
  }
  switch(zero){
  case 0: Zf = 0; break;
  case 1: Zf = 1; break;
  case VF: setZf();
  }
}

void push( char x ){
  M[sp] = x; sp--;
}

char pop(){
  sp++; return M[sp];
}

void push2( int x ){  /* push lo then push hi */
  push( x & 255 );
  push( (x << 8) & 255);
}

int pop2(){  /* pop hi then pop lo */
  int hi,lo;
  hi = pop();
  lo = pop();
  return (int)( (hi << 8) | ( lo & 255) );
}
/* inst: jmp jpz jpc jsr */
int isjmp(int c){
  return ( c == 4 || c == 5 || c == 6 || c == 7 );
}

/* mode 0 IH, 1 IM, 2 RE, 3 DI */
int fetchOperand( int op, int mode ){
  int v, ea;
  switch(mode){
  case 0: v = 0; break;
  case 1: v = M[pc]; pc++; break;  /* IM 8-bit */
  case 2: v = M[pc]; pc++; v += pc; break;  /* RE */
  case 3:
    ea = get2(pc); pc+=2; v = M[ea];  /* DI */
	if( isjmp(op) ) v = ea;
	if( op == STA ) v = ea;
	if( op == LDS ) v = get2(ea);
  }
  return v;
}

/* special trap 1000 print ac, 1001 stop */
void trap(int v){
  switch( v ){
  case PRINT : printf("ac %d\n",ac); break;
  case STOP  : printf("stop\n"); exit(0);
  default    : push2(pc); pc = v; /* jsr */
  }
}

void show(int ip){
  printf("pc:%04X %3s  A:%02X  SP:%04X  C%d Z%d\n",
    ip,optable[op].name,ac&255,sp&0x0ffff,Cf,Zf);
}

/* limit the number of instructions
   to prevent the infinite loop */
void run(int limit)  {
  int pc0, ir, mode, v, c2;

  while(limit--) {
	pc0 = pc;
    ir = M[pc++];
    op = (ir>>4) & 15;
    mode = (ir >> 2) & 3;
    v = fetchOperand(op,mode);
    switch( op ) {
	case LDS: sp = v; setflag(0,1); break;
	case LDA: ac = v; setflag(0,VF); break;
	case PSH: push(ac); setflag(XF,VF); break;
	case STA: M[v] = ac; break;
	case JMP: pc = v; break;
	case JPZ: if( Zf ) pc = v; break;
	case JPC: if( Cf ) pc = v; break;
	case JSR: trap(v); break;
	case AND: ac &= v; setflag(0,VF); break;
	case ORA: ac |= v; setflag(0,VF); break;
	case POP: ac = pop(); setflag(XF,VF); break;
	case XOR: ac ^= v; setflag(0,VF); break;
	case ADD: ac += v; setflag(VF,VF); break;
	case RTS: pc = pop2(); break;
	case ROL: ac = (ac << 1) | Cf; setflag(VF,VF); break;
	case ROR:
		c2 = ac & 1;
       	ac = ac >> 1;
	    if( Cf )  ac |= 0x080;
	    ac &= 255;
	    Cf = c2;
	    setZf();
	    break;
    default: error("undefine instruction");
    }
    if( trace ) show(pc0);
  }
}

int Sreader(char *mem);

void main(){
  int a;
  trace = 1;
  a = Sreader(M);
//  printf("last address loaded %d\n",a);
  pc = 0; Cf = 0; Zf = 0;
  ac = 0; sp = MXMEM-1;
  run(500);
}
